#ifndef _SYS_ATOMIC_H_
#define _SYS_ATOMIC_H_

/*
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log: atomic.h,v $
 * Revision 1.1.1.1  2003/05/09 14:41:18  haraldkipp
 * Initial using 3.2.1
 *
 * Revision 1.4  2003/02/04 18:00:52  harald
 * Version 3 released
 *
 * Revision 1.3  2002/06/26 17:29:28  harald
 * First pre-release with 2.4 stack
 *
 */

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

static inline u_char AtomicInc2(volatile u_char *ptr)
{
    u_char cnt;

    asm volatile(
        "cli"                       "\n\t"
        "ld  %0, %a1"               "\n\t"
        "inc %0"                    "\n\t"
        "st  %a1, %0"               "\n\t"
        : "=&r" (cnt)
        : "e" (ptr)
    );
    return cnt;
}

static inline u_char AtomicDec2(volatile u_char *p)
{
    u_char cnt;

    asm volatile(
        "in  __tmp_reg__, __SREG__" "\n\t"
        "cli"                       "\n\t"
        "ld %0, %a1"                "\n\t"
        "dec %0"                    "\n\t"
        "st %a1, %0"                "\n\t"
        "out __SREG__, __tmp_reg__" "\n\t"
        : "=&r" (cnt)
        : "e" (p)
    );
    return cnt;
}

#ifdef __cplusplus
}
#endif

#endif
